unit ActiveCtrls;
{
 This code was written by  A. Weintrub
 www.weintrub.de
 email: alexander@weintrub.de
+++++++++++++++++++++++++++++++++++++++++++++++++++
+        NUR FR PRIVATE NUTZUNG!!!               +
+++++++++++++++++++++++++++++++++++++++++++++++++++}

interface

uses
  windows, sysutils, graphics, controls, messages, classes, StdCtrls;

type
  TActiveCheckBox = class(TCheckBox)
  private
    FColor, FAColor: TColor;
    IgnoreFontChanged: Boolean;
    FActivateOnEnter: Boolean;
    FActivateOnMouse: Boolean;
    FCaptured: Boolean;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure Activate;
    procedure Deactivate;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property ActiveColor: TColor read FAColor write FAColor;
    property ActivateOnMouse: Boolean read FActivateOnMouse write
      FActivateOnMouse;
    property ActivateOnEnter: Boolean read FActivateOnEnter write
      FActivateOnEnter;
  end;

  TActiveEdit = class(TEdit)
  private
    FColor, FAColor, FBGColor: TColor;
    IgnoreFontChanged: Boolean;
    FActivateOnEnter: Boolean;
    FActivateOnMouse: Boolean;
    FActiveBackGroundColor: TColor;
    FCaptured: Boolean;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure Activate;
    procedure Deactivate;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property ActiveColor: TColor read FAColor write FAColor;
    property ActiveBackGroundColor: TColor read FActiveBackGroundColor write
      FActiveBackGroundColor;
    property ActivateOnMouse: Boolean read FActivateOnMouse write
      FActivateOnMouse;
    property ActivateOnEnter: Boolean read FActivateOnEnter write
      FActivateOnEnter;
  end;

procedure Register;
 {$R ActiveCtrls.dcr}

implementation

procedure Register;
begin
  if SysLocale.PriLangID = LANG_GERMAN then
    RegisterComponents('Beispiele', [TActiveCheckBox, TActiveEdit])
  else
    RegisterComponents('Samples', [TActiveCheckBox, TActiveEdit])
end;

procedure TActiveCheckBox.Activate;
begin
  IgnoreFontChanged := True;
  Font.Color := FAColor;
  IgnoreFontChanged := False;
end;

procedure TActiveCheckBox.Deactivate;
begin
  IgnoreFontChanged := True;
  Font.Color := FColor;
  IgnoreFontChanged := False;
end;

procedure TActiveCheckBox.CMEnter(var Message: TCMEnter);
begin
  if FActivateOnEnter then
    Activate;
  inherited;
end;

procedure TActiveCheckBox.CMExit(var Message: TCMExit);
begin
  if not FCaptured or not FActivateOnMouse then
    Deactivate;
  inherited;
end;

procedure TActiveCheckBox.CMFontChanged(var Message: TMessage);
begin
  inherited;
  if not IgnoreFontChanged then
    FColor := Font.Color;
end;

procedure TActiveCheckBox.CMMouseEnter(var Message: TMessage);
begin
  FCaptured := True;
  if FActivateOnMouse then
    Activate;
  inherited;
end;

procedure TActiveCheckBox.CMMouseLeave(var Message: TMessage);
begin
  FCaptured := False;
  if not FActivateOnEnter or not Focused then
    Deactivate;
  inherited;
end;

constructor TActiveCheckBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  IgnoreFontChanged := False;
  FActivateOnEnter := True;
  FActivateOnMouse := True;
  FColor := Font.Color;
  FAColor := clRed;
end;

{ TActiveEdit }

procedure TActiveEdit.Activate;
begin
  IgnoreFontChanged := True;
  Font.Color := FAColor;
  if FActivateOnEnter and Focused  then
    Color := FActiveBackGroundColor;
  IgnoreFontChanged := False;
end;

procedure TActiveEdit.CMColorChanged(var Message: TMessage);
begin
  inherited;
  if not IgnoreFontChanged then
    FBGColor := Color;
end;

procedure TActiveEdit.CMEnter(var Message: TCMEnter);
begin
  if FActivateOnEnter then
    Activate;
  inherited;
end;

procedure TActiveEdit.CMExit(var Message: TCMExit);
begin
  if not FCaptured or not FActivateOnMouse then
    Deactivate;
  inherited;
end;

procedure TActiveEdit.CMFontChanged(var Message: TMessage);
begin
  inherited;
  if not IgnoreFontChanged then
    FColor := Font.Color;
end;

procedure TActiveEdit.CMMouseEnter(var Message: TMessage);
begin
  FCaptured := True;
  if FActivateOnMouse then
    Activate;
  inherited;
end;

procedure TActiveEdit.CMMouseLeave(var Message: TMessage);
begin
  FCaptured := False;
  if not FActivateOnEnter or not Focused then
    Deactivate;
  inherited;
end;

constructor TActiveEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  IgnoreFontChanged := False;
  FActivateOnEnter := True;
  FActivateOnMouse := True;
  FAColor := clRed;
  FColor := Font.Color;
  FActiveBackGroundColor := Color;
  FBGColor := clBtnFace;
  Deactivate;
end;

procedure TActiveEdit.Deactivate;
begin
  IgnoreFontChanged := True;
  Font.Color := FColor;
  Color := FBGColor;
  IgnoreFontChanged := False;
end;

end.

